/* Copyright 2012 Marco Minerva, marco.minerva@gmail.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using Microsoft.SPOT;
using Gadgeteer.Modules.GHIElectronics;

namespace GadgeteerRobot
{
    public class MovementControl
    {
        private MotorControllerL298 motorController;

        public enum RoverMovement
        {
            Idle, MoveForward, MoveBackward, TurnRight, TurnLeft, Rotate
        }

        public enum RoverDirection
        {
            None, Forward, Backward
        }

        private const int DEFAULT_SPEED = 50;

        public bool Started { get; private set; }
        public RoverMovement CurrentMovement { get; private set; }
        public RoverDirection CurrentDirection { get; private set; }
        public int CurrentSpeed { get; private set; }
                
        private RoverMovement newMovement;
        private RoverDirection newDirection;
        private int newSpeed;

        public MovementControl(MotorControllerL298 motorController)
        {
            this.motorController = motorController;

            Started = false;
            CurrentSpeed = newSpeed = DEFAULT_SPEED;
            CurrentDirection = newDirection = RoverDirection.None;
            CurrentMovement = newMovement = RoverMovement.Idle;
        }

        public void TurnOn()
        {
            Started = true;
        }

        public void TurnOff()
        {
            Started = false;

            // Imposta i parametri in modo tale che il metodo UpdateRobotStatus fermi il movimento.
            newDirection = RoverDirection.None;
            newMovement = RoverMovement.Idle;
            newSpeed = DEFAULT_SPEED;
            this.UpdateRobotStatus();
        }

        public void SetSpeed(int newSpeed)
        {
            this.newSpeed = newSpeed;
            this.UpdateRobotStatus();
        }

        public void Move(RoverMovement newMovement, RoverDirection newDirection)
        {
            this.newMovement = newMovement;
            this.newDirection = newDirection;
            this.UpdateRobotStatus();
        }

        private void UpdateRobotStatus()
        {
            if (CurrentMovement != newMovement
                || CurrentDirection != newDirection
                || CurrentSpeed != newSpeed)
            {
                // Imposta velocit e direzione dei motori in base ai comandi ricevuti.
                CurrentMovement = newMovement;
                CurrentDirection = newDirection;
                CurrentSpeed = newSpeed;

                int speed = CurrentSpeed;
                int motor1Direction = 0;
                int motor2Direction = 0;

                switch (CurrentMovement)
                {
                    case RoverMovement.Idle:
                        motor1Direction = motor2Direction = 0;
                        break;

                    case RoverMovement.MoveForward:
                        motor1Direction = motor2Direction = 1;
                        break;

                    case RoverMovement.MoveBackward:
                        motor1Direction = motor2Direction = -1;
                        break;

                    case RoverMovement.TurnRight:
                        speed = System.Math.Min(speed * 2, 100);
                        motor1Direction = 0;
                        if (CurrentDirection == RoverDirection.Forward)
                            motor2Direction = 1;
                        else if (CurrentDirection == RoverDirection.Backward)
                            motor2Direction = -1;
                        break;

                    case RoverMovement.TurnLeft:
                        speed = System.Math.Min(speed * 2, 100);
                        motor2Direction = 0;
                        if (CurrentDirection == RoverDirection.Forward)
                            motor1Direction = 1;
                        else if (CurrentDirection == RoverDirection.Backward)
                            motor1Direction = -1;
                        break;

                    case RoverMovement.Rotate:
                        speed = System.Math.Min(speed * 2, 100);
                        motor1Direction = 1;
                        motor2Direction = -1;
                        break;

                    default:
                        break;
                }

                motorController.MoveMotor(MotorControllerL298.Motor.Motor1, speed * motor1Direction);
                motorController.MoveMotor(MotorControllerL298.Motor.Motor2, speed * motor2Direction);
            }
        }
    }
}
