﻿using System;
using System.Collections;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Presentation;
using Microsoft.SPOT.Presentation.Controls;
using Microsoft.SPOT.Presentation.Media;
using Microsoft.SPOT.Touch;
using System.IO;

using Gadgeteer.Networking;
using GT = Gadgeteer;
using GTM = Gadgeteer.Modules;
using Gadgeteer.Modules.GHIElectronics;
using NetUtils;

namespace GadgeteerRobot
{
    public partial class Program
    {
        private const string CONFIG_FILE_PATH = "GadgeteerRobot.config";

        private string networkSsid;
        private string passphrase;
        private string netbiosName;

        private NameService nameService;

        private TcpSocketServer movementServer;
        private TcpSocketServer cameraServer;

        private MovementControl movementControl;

        private byte[] currentBitmapData;

        private bool started;

        private static object syncObject = new object();

        // This method is run when the mainboard is powered up or reset.   
        void ProgramStarted()
        {
            led.GreenBlueSwapped = true;
            led.TurnWhite();

            button.ButtonPressed += new Button.ButtonEventHandler(button_ButtonPressed);
            wifi.NetworkUp += new GTM.Module.NetworkModule.NetworkEventHandler(wifi_NetworkUp);
            wifi.UseDHCP();

            movementServer = new TcpSocketServer(8080);
            movementServer.DataReceived += new TcpSocketServer.DataReceivedEventHandler(movementServer_DataReceived);

            cameraServer = new TcpSocketServer(8081);
            cameraServer.DataReceived += new TcpSocketServer.DataReceivedEventHandler(cameraServer_DataReceived);

            camera.CurrentPictureResolution = Camera.PictureResolution.Resolution160x120;
            camera.BitmapStreamed += new Camera.BitmapStreamedEventHandler(camera_BitmapStreamed);

            movementControl = new MovementControl(motorControllerL298);

            // Use Debug.Print to show messages in Visual Studio's "Output" window during debugging.
            Debug.Print("Program Started");
        }

        private void button_ButtonPressed(Button sender, Button.ButtonState state)
        {
            if (started)
                return;

            started = true;
            this.ReadConfiguration();
        }

        private void ReadConfiguration()
        {
            if (sdCard.IsCardMounted)
            {
                led.BlinkRepeatedly(GT.Color.Blue);
                try
                {
                    // Reads configuration file.
                    using (Stream configStream =
                        sdCard.GetStorageDevice().OpenRead(CONFIG_FILE_PATH))
                    {
                        // Loads settings.
                        ConfigurationManager.Load(configStream);
                    }

                    // Retrieves configuration values.
                    networkSsid = ConfigurationManager.GetAppSetting("NetworkSsid");
                    passphrase = ConfigurationManager.GetAppSetting("NetworkPassphrase");
                    netbiosName = ConfigurationManager.GetAppSetting("NetbiosName");

                    var info = new WiFi_RS21.WiFiNetworkInfo();
                    info.SSID = networkSsid;
                    info.SecMode = WiFi_RS21.SecurityMode.WPA2;
                    info.networkType = WiFi_RS21.NetworkType.AccessPoint;

                    wifi.Join(info, passphrase);
                }
                catch
                {
                    led.SetGreenIntensity(255);
                    led.SetRedIntensity(255);
                }
            }
            else
            {
                led.SetGreenIntensity(255);
                led.SetRedIntensity(255);
            }
        }

        private void wifi_NetworkUp(GTM.Module.NetworkModule sender, GTM.Module.NetworkModule.NetworkState state)
        {
            Debug.Print("IP Address: " + wifi.NetworkSettings.IPAddress);

            // Registra il nome del device con NetBios.
            nameService = new NameService();
            nameService.AddName(netbiosName, NameService.NameType.Unique, NameService.MsSuffix.Default);

            camera.StartStreamingBitmaps(
                new Bitmap(camera.CurrentPictureResolution.Width,
                camera.CurrentPictureResolution.Height));

            movementServer.Start();
            cameraServer.Start();

            led.TurnRed();
        }

        private void movementServer_DataReceived(object sender, TcpDataReceivedEventArgs e)
        {
            string data = TcpSocketServer.BytesToString(e.Data).ToLower();
            Debug.Print("Received data: " + data);

            if (data == RobotCommands.TURN_ON)
            {
                // Accende i motori, permettendo il movimento del Robot.
                movementControl.TurnOn();

                led.TurnGreen();
            }
            else if (data == RobotCommands.TURN_OFF && movementControl.Started)
            {
                // Spegne il motore, quindi annulla anche i movimenti.
                movementControl.TurnOff();

                led.TurnRed();
            }
            else if (movementControl.Started)
            {
                if (data.IndexOf(RobotCommands.SET_SPEED) >= 0)
                {
                    // E' stato ricevuto il comando per impostare una nuova velocità.
                    try
                    {
                        int newSpeed = int.Parse(data.Substring(data.LastIndexOf(":") + 1));
                        movementControl.SetSpeed(newSpeed);
                    }
                    catch { }
                }
                else
                {
                    // Sono stati ricevuti comandi relativi al movimento del Robot. Li processa.
                    this.HandleMovementCommands(data);
                }
            }
        }

        private void HandleMovementCommands(string data)
        {
            var newMovement = movementControl.CurrentMovement;
            var newDirection = movementControl.CurrentDirection;

            foreach (char command in data)
            {
                switch (command.ToString())
                {
                    case RobotCommands.MOVE_FORWARD:
                        newMovement = MovementControl.RoverMovement.MoveForward;
                        newDirection = MovementControl.RoverDirection.Forward;
                        break;

                    case RobotCommands.MOVE_BACKWARD:
                        newMovement = MovementControl.RoverMovement.MoveBackward;
                        newDirection = MovementControl.RoverDirection.Backward;
                        break;

                    case RobotCommands.TURN_RIGHT:
                        newMovement = MovementControl.RoverMovement.TurnRight;
                        break;

                    case RobotCommands.TURN_LEFT:
                        newMovement = MovementControl.RoverMovement.TurnLeft;
                        break;

                    case RobotCommands.ROTATE:
                        newMovement = MovementControl.RoverMovement.Rotate;
                        break;

                    case RobotCommands.STOP:
                        newMovement = MovementControl.RoverMovement.Idle;
                        break;

                    default:
                        break;
                }
            }

            movementControl.Move(newMovement, newDirection);
        }

        private void camera_BitmapStreamed(Camera sender, Bitmap bitmap)
        {
            lock (syncObject)
            {
                currentBitmapData = new byte[bitmap.Width * bitmap.Height * 3 + 54];
                GHIElectronics.NETMF.System.Util.BitmapToBMPFile(bitmap.GetBitmap(),
                                                                 bitmap.Width,
                                                                 bitmap.Height, currentBitmapData);
            }
        }

        private void cameraServer_DataReceived(object sender, TcpDataReceivedEventArgs e)
        {
            string data = TcpSocketServer.BytesToString(e.Data).ToLower();
            Debug.Print("Received camera data: " + data);

            if (data == RobotCommands.GET_IMAGE && currentBitmapData != null)
            {
                lock (syncObject)
                {
                    // Sends the current image to the client.
                    e.ResponseData = currentBitmapData;
                }
            }
            //else if (data == RobotCommands.CENTER_CAMERA)
            //{
            //    servoControl.Center();
            //}
            //else if (data.IndexOf(RobotCommands.CAMERA_LEFT) >= 0)
            //{
            //    // E' stato ricevuto il comando per ruotare la camera a sinistra.
            //    try
            //    {
            //        int step = int.Parse(data.Substring(data.LastIndexOf(":") + 1));
            //        servoControl.Move(ServoDirection.Left, step);
            //    }
            //    catch { }
            //}
            //else if (data.IndexOf(RobotCommands.CAMERA_RIGHT) >= 0)
            //{
            //    // E' stato ricevuto il comando per ruotare la camera a destra.
            //    try
            //    {
            //        int step = int.Parse(data.Substring(data.LastIndexOf(":") + 1));
            //        servoControl.Move(ServoDirection.Right, step);
            //    }
            //    catch { }
            //}
        }
    }
}
