﻿/* Copyright 2012 Marco Minerva, marco.minerva@gmail.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;
using System.Net.Sockets;
using GadgeteerManager;
using System.Threading;

namespace GadgeteerControlLibrary
{
    public class CameraControl : ICameraControl
    {
        private TcpClient client;
        private Stream tcpStream;
        private Timer tmrPing;

        private const int IMAGE_SIZE = 57654;

        public void Connect(string address, string password)
        {
            if (string.IsNullOrWhiteSpace(address))
                throw new ArgumentNullException(address);

            try
            {
                string[] addressPart = address.Split(':');
                client = new TcpClient();
                client.Connect(addressPart[0], int.Parse(addressPart[1]));
                tcpStream = client.GetStream();

                tmrPing = new Timer((stateInfo) => Ping(), null,
                    (int)TimeSpan.FromSeconds(30).TotalMilliseconds, (int)TimeSpan.FromSeconds(30).TotalMilliseconds);
            }
            catch
            {
                throw new ConnectionException("Rover not found or connection error");
            }
        }

        public void Disconnect()
        {
            try
            {
                tmrPing.Dispose();
                tmrPing = null;

                tcpStream.Close();
                tcpStream.Dispose();
                tcpStream = null;
                client.Close();
            }
            catch { }
        }

        public void Ping()
        {
            this.SendCommand(RobotCommands.PING);
        }

        public byte[] GetImage()
        {
            this.SendCommand(RobotCommands.GET_IMAGE);
            using (MemoryStream ms = new MemoryStream())
            {
                byte[] buffer = new byte[IMAGE_SIZE];
                int bytesRead, totalBytesRead = 0;
                while (totalBytesRead < IMAGE_SIZE)
                {
                    bytesRead = tcpStream.Read(buffer, 0, buffer.Length);

                    totalBytesRead += bytesRead;
                    ms.Write(buffer, 0, bytesRead);
                }

                return ms.ToArray();
            }
        }

        public void Center()
        {
            this.SendCommand(RobotCommands.CENTER_CAMERA);
        }

        public void Move(CameraDirection cameraDirection, int step)
        {
            string command = null;
            switch (cameraDirection)
            {
                case CameraDirection.Left:
                    command = RobotCommands.CAMERA_LEFT;
                    break;

                case CameraDirection.Right:
                    command = RobotCommands.CAMERA_RIGHT;
                    break;

                default:
                    break;
            }

            if (!string.IsNullOrEmpty(command))
            {
                command = string.Format("{0}:{1}", command, step);
                this.SendCommand(command);
            }
        }

        private void SendCommand(string command)
        {
            Debug.WriteLine("Sending command: " + command);
            if (tcpStream != null)
            {
                var buffer = System.Text.Encoding.Default.GetBytes(command);
                tcpStream.Write(buffer, 0, buffer.Length);
            }
        }
    }
}
