﻿/* Copyright 2012 Marco Minerva, marco.minerva@gmail.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;
using System.Net.Sockets;
using GadgeteerManager;
using System.Threading;

namespace GadgeteerControlLibrary
{
    public class RobotControl : IRobotControl, IDisposable
    {
        private TcpClient client;
        private Stream tcpStream;
        private Timer tmrPing;
        private string lastSentCommand;

        public void Connect(string address, string password)
        {
            if (string.IsNullOrWhiteSpace(address))
                throw new ArgumentNullException(address);

            try
            {
                string[] addressPart = address.Split(':');
                client = new TcpClient();
                client.Connect(addressPart[0], int.Parse(addressPart[1]));
                tcpStream = client.GetStream();

                tmrPing = new Timer((stateInfo) => Ping(), null, 
                    (int)TimeSpan.FromSeconds(30).TotalMilliseconds, (int)TimeSpan.FromSeconds(30).TotalMilliseconds);
            }
            catch
            {
                throw new ConnectionException("Rover not found or connection error");
            }
        }

        public void Disconnect()
        {
            try
            {
                tmrPing.Dispose();
                tmrPing = null;

                tcpStream.Close();
                tcpStream.Dispose();
                tcpStream = null;
                client.Close();
            }
            catch { }
        }

        public void Ping()
        {
            this.SendCommand(RobotCommands.PING);
        }

        public void MotorOn()
        {
            this.SendCommand(RobotCommands.TURN_ON);
        }

        public void MotorOff()
        {
            this.SendCommand(RobotCommands.TURN_OFF);
        }

        public void ChangeSpeed(int speed)
        {
            var command = string.Format("{0}:{1}", RobotCommands.SET_SPEED, speed);
            this.SendCommand(command);
        }

        public void MoveForeward()
        {
            this.SendCommand(RobotCommands.MOVE_FORWARD);
        }

        public void MoveBackward()
        {
            this.SendCommand(RobotCommands.MOVE_BACKWARD);
        }

        public void TurnLeft()
        {
            this.SendCommand(RobotCommands.TURN_LEFT);
        }

        public void TurnRight()
        {
            this.SendCommand(RobotCommands.TURN_RIGHT);
        }

        public void MoveForewardLeft()
        {
            this.SendCommand(RobotCommands.MOVE_FORWARD + RobotCommands.TURN_LEFT);
        }

        public void MoveForewardRight()
        {
            this.SendCommand(RobotCommands.MOVE_FORWARD + RobotCommands.TURN_RIGHT);
        }

        public void MoveBackwardLeft()
        {
            this.SendCommand(RobotCommands.MOVE_BACKWARD + RobotCommands.TURN_LEFT);
        }

        public void MoveBackwardRight()
        {
            this.SendCommand(RobotCommands.MOVE_BACKWARD + RobotCommands.TURN_RIGHT);
        }

        public void Stop()
        {
            this.SendCommand(RobotCommands.STOP);
        }

        public void Rotate()
        {
            this.SendCommand(RobotCommands.ROTATE);
        }

        public void Dispose()
        {
            this.Disconnect();
        }

        private void SendCommand(string command)
        {
            if (command != lastSentCommand)
            {
                Debug.WriteLine("Sending command: " + command);
                if (tcpStream != null)
                {
                    var buffer = System.Text.Encoding.Default.GetBytes(command);
                    tcpStream.Write(buffer, 0, buffer.Length);
                }

                lastSentCommand = command;
            }
        }
    }
}
