﻿Imports KinectManager.Core.EventArgs
Imports Microsoft.Kinect

Namespace Handlers
    ''' <summary>
    ''' Class GestureHandlerBase
    ''' </summary>
    Public MustInherit Class GestureHandlerBase
        Inherits KinectHandlerBase

        ''' <summary>
        ''' Gets or sets the minimum delay for gesture (in milliseconds).
        ''' </summary>
        ''' <value>The minimum delay for gesture.</value>
        Public Property MinimumDelayForGesture As UInt32 = 250

        ''' <summary>
        ''' Gets or sets the last gesture signal.
        ''' </summary>
        ''' <value>The last gesture signal.</value>
        Protected Property LastGestureSignal As DateTime

        Public MustOverride Property Name As String

        ''' <summary>
        ''' Elaborates this instance.
        ''' </summary>
        Public MustOverride Sub Elaborate(ByRef stopElaborate As Boolean)

        Protected Function GetTrackedSkeletons() As IEnumerable(Of Skeleton)
            If Me.KinectManager IsNot Nothing Then
                Dim trackedListId = Me.KinectManager.TrackingIdList
                If trackedListId IsNot Nothing AndAlso trackedListId.Any() AndAlso Me.GetLastFrame().Skeletons IsNot Nothing Then
                    Dim trackedSkeletons = Me.GetLastFrame().Skeletons.Where(Function(a) trackedListId.Contains(a.TrackingId))
                    Return trackedSkeletons
                End If
            End If
            Return Nothing
        End Function

        ''' <summary>
        ''' Gets the last frame.
        ''' </summary>
        ''' <returns></returns>
        Protected Function GetLastFrame() As KinectFrameData
            If Me.KinectManager IsNot Nothing Then
                If Me.KinectManager.FrameBuffer.Any() Then
                    Return Me.KinectManager.FrameBuffer.OrderByDescending(Function(k) k.TimeStamp).First()
                Else
                    Return Nothing
                End If
            Else
                Return Nothing
            End If
        End Function

        ''' <summary>
        ''' Gets the last frames by timespan.
        ''' </summary>
        ''' <param name="timespan">The timespan.</param>
        ''' <returns></returns>
        Protected Function GetLastFramesByTimespan(timespan As TimeSpan) As IEnumerable(Of KinectFrameData)
            If Me.KinectManager IsNot Nothing Then
                If Me.KinectManager.FrameBuffer.Any() Then
                    Dim frames = (From s In Me.KinectManager.FrameBuffer
                            Where s.TimeStamp.Add(timespan) > DateTime.Now
                            Select s)
                    Return frames
                Else
                    Return Nothing
                End If
            Else
                Return Nothing
            End If
        End Function

        ''' <summary>
        ''' Gets the last skeletons by timespan and trackid.
        ''' </summary>
        ''' <param name="timespan">The timespan.</param>
        ''' <param name="trackId">The track id.</param>
        ''' <returns></returns>
        Protected Function GetLastSkeletonsByTimespanAndTrackid(timespan As TimeSpan, trackId As Integer) As IEnumerable(Of Skeleton)
            If Me.KinectManager IsNot Nothing Then
                If Me.KinectManager.FrameBuffer.Any() Then

                    Dim frames = GetLastFramesByTimespan(timespan)
                    Return From f In frames
                        Select (From s In f.Skeletons
                            Where s.TrackingId = trackId
                            Select s).FirstOrDefault()

                Else
                    Return Nothing
                End If
            Else
                Return Nothing
            End If
        End Function

        ''' <summary>
        ''' Raises the <see cref="E:GestureRecognized" /> event.
        ''' </summary>
        ''' <param name="e">The <see cref="GestureRecognizedEventArgs" /> instance containing the event data.</param>
        Protected Sub OnGestureRecognized(e As GestureRecognizedEventArgs)
            If Me.MinimumDelayForGesture = 0 Or Me.LastGestureSignal.AddMilliseconds(CDbl(Me.MinimumDelayForGesture)) <= DateTime.Now Then
                Me.LastGestureSignal = DateTime.Now
                RaiseEvent GestureRecognized(Me, e)
            End If
        End Sub

        ''' <summary>
        ''' Occurs when [gesture recognized].
        ''' </summary>
        Public Event GestureRecognized(sender As Object, e As GestureRecognizedEventArgs)
    End Class
End Namespace