﻿Imports System.Collections.Concurrent
Imports System.IO
Imports Microsoft.Speech.AudioFormat
Imports Microsoft.Kinect
Imports KinectManager.Core.Handlers
Imports KinectManager.Core.EventArgs
Imports System.Threading.Tasks
Imports System.Threading
Imports Microsoft.Speech.Recognition

Public Class KinectManager
    Implements IDisposable

    Public Sub New(colorImageFormat As ColorImageFormat, depthImageFormat As DepthImageFormat)
        Me.ColorImageFormat = colorImageFormat
        Me.DepthImageFormat = depthImageFormat

        Me._GestureHandlers = New List(Of GestureHandlerBase)()
        Me._FrameBuffer = New List(Of KinectFrameData)()
    End Sub


#Region "Properties"

    Public Property SensorElevationAngle As Integer
        Get
            If Me._Sensor IsNot Nothing Then
                Return Me._Sensor.ElevationAngle
            End If
            Return Nothing
        End Get
        Set(value As Integer)
            If Me._Sensor IsNot Nothing Then
                Me._Sensor.ElevationAngle = value
            End If
        End Set
    End Property

    Public ReadOnly Property ReferenceAreas As IEnumerable(Of ReferenceArea)
        Get
            Dim retList As New List(Of ReferenceArea)
            If Me.SkeletalTrackingHandler IsNot Nothing AndAlso Me.SkeletalTrackingHandler.ReferenceAreas.Any() Then
                retList.AddRange(Me.SkeletalTrackingHandler.ReferenceAreas)
            End If
            If Me.GestureHandlers IsNot Nothing Then
                For Each gh In Me.GestureHandlers
                    If gh.ReferenceAreas.Any Then
                        retList.AddRange(gh.ReferenceAreas)
                    End If
                Next
            End If
            Return retList
        End Get
    End Property

    Public ReadOnly Property Status As KinectStatus
        Get
            If Me._Sensor IsNot Nothing Then
                Return _Sensor.Status
            Else
                Return KinectStatus.Undefined
            End If
        End Get
    End Property

    Private _ColorImageFormat As ColorImageFormat
    Public Property ColorImageFormat As ColorImageFormat
        Get
            Return _ColorImageFormat
        End Get
        Protected Set(value As ColorImageFormat)
            _ColorImageFormat = value
        End Set
    End Property

    Private _DepthImageFormat As DepthImageFormat
    Public Property DepthImageFormat As DepthImageFormat
        Get
            Return _DepthImageFormat
        End Get
        Protected Set(value As DepthImageFormat)
            _DepthImageFormat = value
        End Set
    End Property

    Public ReadOnly Property CurrentColorImageFrame As ColorImageFrame
        Get
            If Me._FrameBuffer IsNot Nothing AndAlso Me._FrameBuffer.Any() Then
                Return Me._FrameBuffer.Last().ColorFrame
            End If
            Return Nothing
        End Get
    End Property
    
    Public ReadOnly Property CurrentDepthImageFrame As DepthImageFrame
        Get
            If Me._FrameBuffer IsNot Nothing AndAlso Me._FrameBuffer.Any() Then
                Return Me._FrameBuffer.Last().DepthFrame
            End If
            Return Nothing
        End Get
    End Property


    Public ReadOnly Property CurrentSkeletons As IEnumerable(Of Skeleton)
        Get
            If Me._FrameBuffer IsNot Nothing AndAlso Me._FrameBuffer.Any() Then
                Return Me._FrameBuffer.Last().Skeletons
            Else
                Return Nothing
            End If
        End Get
    End Property

    Public Property Mode As KinectManagerMode

    Public Property Features As KinectManagerFeature
#End Region

#Region "KinectSensor"
    Private _Sensor As KinectSensor = Nothing

    Private Sub StartKinectSensor(mode As KinectManagerMode, features As KinectManagerFeature)
        Me.Mode = mode
        Me.Features = features
        _Sensor = KinectSensor.KinectSensors.FirstOrDefault(Function(s) s.Status = KinectStatus.Connected)
        If _Sensor IsNot Nothing AndAlso Not _Sensor.IsRunning Then
            If (Me.Features And KinectManagerFeature.Gesture) > 0 Then
                Select Case Me.Mode
                    Case KinectManagerMode.EventDriven
                        AddHandler _Sensor.AllFramesReady, AddressOf AllFrameReadyHandler
                    Case KinectManagerMode.Polling
                        Me._PollingCancellationTokenSource = New CancellationTokenSource
                        _PollingCancellationToken = Me._PollingCancellationTokenSource.Token
                        Dim taskFactory = New TaskFactory()
                        Me._PollingTask = taskFactory.StartNew(AddressOf PollingTaskCode)
                End Select
                _Sensor.ColorStream.Enable(ColorImageFormat)
                _Sensor.DepthStream.Enable(DepthImageFormat)
                _Sensor.SkeletonStream.Enable()                
            End If
            Try
                _Sensor.Start()
            Catch ex As Exception
                Throw
            End Try
            If (Me.Features And KinectManagerFeature.Speech) > 0 Then
                StartSpeechRecognition(SpeechRecognitionGrammarFileName)
            End If            
        End If
    End Sub

    Private _PollingTask As task
    Private _PollingCancellationTokenSource As CancellationTokenSource
    Private _PollingCancellationToken As CancellationToken

    Private Sub StopKinectSensor()
        If _Sensor IsNot Nothing Then
            If (Me.Features And KinectManagerFeature.Gesture) > 0 Then
                Select Case Me.Mode
                    Case KinectManagerMode.EventDriven
                        RemoveHandler _Sensor.AllFramesReady, AddressOf AllFrameReadyHandler
                    Case KinectManagerMode.Polling
                        Me._PollingCancellationTokenSource.Cancel()
                        Me._PollingTask = Nothing
                End Select
            End If
            If (Me.Features And KinectManagerFeature.Speech) > 0 Then
                StopSpeechRecognition()
            End If
            If _Sensor.IsRunning Then
                _Sensor.Stop()
            End If
        End If
    End Sub

    Private Sub AllFrameReadyHandler(sender As Object, e As AllFramesReadyEventArgs)
        Dim colorImageFrame = e.OpenColorImageFrame()
        Dim depthImageFrame = e.OpenDepthImageFrame()
        Dim skeletonFrame = e.OpenSkeletonFrame
        ElaborateFrames(colorImageFrame, depthImageFrame, skeletonFrame)
    End Sub

    Private Sub PollingTaskCode()
        While (Not Me._PollingCancellationToken.IsCancellationRequested)
            If Me._Sensor.IsRunning Then
                Dim colorImageFrame = Me._Sensor.ColorStream.OpenNextFrame(0)
                Dim depthImageFrame = Me._Sensor.DepthStream.OpenNextFrame(0)
                Dim skeletonFrame = Me._Sensor.SkeletonStream.OpenNextFrame(0)
                ElaborateFrames(colorImageFrame, depthImageFrame, skeletonFrame)
            End If
            Thread.Sleep(50)
        End While
    End Sub

    Protected Sub ElaborateFrames(ByVal colorImageFrame As ColorImageFrame, ByVal depthImageFrame As DepthImageFrame, ByVal skeletonFrame As SkeletonFrame)
        Dim frameData = CreateFrameData(colorImageFrame, depthImageFrame, skeletonFrame)
        AddFrameDataToFrameBuffer(frameData)
        ElaborateSkeletalTracking()
        NotifyToGestureHandlers()
        RaiseEvent FrameChanged(Me, System.EventArgs.Empty)
    End Sub

    Private Function CreateFrameData(colorImageFrame As ColorImageFrame,
                                     depthImageFrame As DepthImageFrame,
                                     skeletonFrame As SkeletonFrame) As KinectFrameData
        Dim frame = New KinectFrameData
        frame.ColorFrame = colorImageFrame
        frame.DepthFrame = depthImageFrame        
        If skeletonFrame IsNot Nothing Then
            frame.Skeletons = CreateSkeletonsData(skeletonFrame)
        End If
        Return frame
    End Function

    Private Function CreateColorImageData(colorImageFrame As ColorImageFrame) As IEnumerable(Of Byte)
        If colorImageFrame IsNot Nothing Then
            Dim imageArray() = New Byte(colorImageFrame.PixelDataLength - 1) {}
            colorImageFrame.CopyPixelDataTo(imageArray)
            Return imageArray
        Else
            Return Nothing
        End If
    End Function

    Private Function CreateDepthImageData(depthImageFrame As DepthImageFrame) As IEnumerable(Of Short)
        If depthImageFrame IsNot Nothing Then
            Dim depthArray() = New Short(depthImageFrame.PixelDataLength - 1) {}
            depthImageFrame.CopyPixelDataTo(depthArray)
            Return depthArray
        Else
            Return Nothing
        End If
    End Function

    Private Function CreateSkeletonsData(skeletonFrame As SkeletonFrame) As IEnumerable(Of Skeleton)
        If skeletonFrame IsNot Nothing Then
            Dim skeletons() As Skeleton = New Skeleton(5) {}
            skeletonFrame.CopySkeletonDataTo(skeletons)
            Return skeletons
        Else
            Return Nothing
        End If
    End Function
#End Region

#Region "Gesture Handlers"
    Private Sub GestureRecognizedHandler(sender As Object, e As GestureRecognizedEventArgs)
        RaiseEvent GestureRecognized(Me, e)
    End Sub

    Public Sub AddGestureHandler(gestureHandler As GestureHandlerBase)
        If gestureHandler Is Nothing Then Throw New ArgumentNullException("gestureHandler")
        gestureHandler.KinectManager = Me
        AddHandler gestureHandler.GestureRecognized, AddressOf GestureRecognizedHandler
        Me._GestureHandlers.Add(gestureHandler)
    End Sub

    Public Function RemoveGestureHandler(gestureHandler As GestureHandlerBase) As Boolean
        If gestureHandler Is Nothing Then Throw New ArgumentNullException("gestureHandler")
        If Me._GestureHandlers.Remove(gestureHandler) Then
            RemoveHandler gestureHandler.GestureRecognized, AddressOf GestureRecognizedHandler
            gestureHandler.KinectManager = Nothing
            Return True
        Else
            Return False
        End If
    End Function

    Private _GestureHandlers As List(Of GestureHandlerBase)
    Public ReadOnly Property GestureHandlers As IEnumerable(Of GestureHandlerBase)
        Get
            Return _GestureHandlers
        End Get
    End Property

    Private Sub NotifyToGestureHandlers()
        Dim stopElaborate As Boolean = False
        For Each g In Me._GestureHandlers
            g.Elaborate(stopElaborate)
            If stopElaborate Then Exit For
        Next
    End Sub
#End Region

#Region "Skeletal tracking Handlers"

    Private _SkeletalTrackingHandler As SkeletalTrackingHandlerBase
    Public Property SkeletalTrackingHandler As SkeletalTrackingHandlerBase
        Get
            Return _SkeletalTrackingHandler
        End Get
        Set(value As SkeletalTrackingHandlerBase)
            If Me._SkeletalTrackingHandler IsNot Nothing Then
                RemoveHandler Me._SkeletalTrackingHandler.SkeletalTracked, AddressOf SkeletalTrackedHandler
            End If
            Me._SkeletalTrackingHandler = value
            If value IsNot Nothing Then
                Me._SkeletalTrackingHandler.KinectManager = Me
                AddHandler Me._SkeletalTrackingHandler.SkeletalTracked, AddressOf SkeletalTrackedHandler
            End If
        End Set
    End Property

    Private Sub ElaborateSkeletalTracking()
        If SkeletalTrackingHandler Is Nothing Then
            If Me.FrameBuffer.Any() Then
                TrackingIdList = (From i In Me.FrameBuffer.Last().Skeletons
                                      Where i.TrackingState = SkeletonTrackingState.Tracked
                                      Select i.TrackingId).Take(1)

                If TrackingIdList.Any() Then
                    For Each id In TrackingIdList
                        Dim skeleton = (From i In Me.FrameBuffer.Last().Skeletons
                                      Where i.TrackingId = id
                                      Select i).FirstOrDefault()

                        Dim arg = New SkeletalTrackedEventArgs() With {.Timestamp = DateTime.Now, .Skeleton = skeleton}
                        RaiseEvent SkeletalTracked(Me, arg)
                    Next
                End If
            End If
        Else
            TrackingIdList = SkeletalTrackingHandler.Elaborate()
        End If
    End Sub

    Private TrackingIdListSyncObject As Object = New Object

    Private _TrackingIdList As IEnumerable(Of Integer)
    Public Property TrackingIdList As IEnumerable(Of Integer)
        Get
            Return _TrackingIdList
        End Get
        Protected Set(value As IEnumerable(Of Integer))
            SyncLock TrackingIdListSyncObject
                _TrackingIdList = value
            End SyncLock
        End Set
    End Property

    Private Sub SkeletalTrackedHandler(sender As Object, e As SkeletalTrackedEventArgs)
        RaiseEvent SkeletalTracked(Me, e)
    End Sub

#End Region

#Region "FrameData Buffer"
    Private Sub AddFrameDataToFrameBuffer(frameData As KinectFrameData)
        If frameData Is Nothing Then Throw New ArgumentNullException("frameData")
        SyncLock FrameBufferSyncObject
            Me._FrameBuffer.Add(frameData)
            Me._FrameBuffer.RemoveAll(Function(a) a.TimeStamp.Add(Me.FrameBufferDuration) < frameData.TimeStamp)
        End SyncLock
    End Sub


    Private FrameBufferSyncObject As Object = New Object

    Private _FrameBuffer As List(Of KinectFrameData)
    Public ReadOnly Property FrameBuffer As IEnumerable(Of KinectFrameData)
        Get
            Return _FrameBuffer
        End Get
    End Property

    Public Property FrameBufferDuration As TimeSpan = New TimeSpan(0, 0, 5)

#End Region

#Region "Speech Recognition"
    Private _SpeechEngine As SpeechRecognitionEngine

    Public Property SpeechRecognitionLanguage As String = "en-US"
    Public Property SpeechRecognitionGrammarFileName As String = ""

    Public Sub StartSpeechRecognition(grammarFileName As String)
        If String.IsNullOrWhiteSpace(grammarFileName) Then Throw New ArgumentNullException()
        If Not File.Exists(grammarFileName) Then Throw New FileNotFoundException
        If Me._Sensor.IsRunning Then
            Dim grammar = New Grammar(grammarFileName)
            SetSpeechEngine(grammar)
        End If
    End Sub

    Public Sub StartSpeechRecognition(grammar As Grammar)
        If grammar Is Nothing Then Throw New ArgumentNullException()
        If Me._Sensor.IsRunning Then
            SetSpeechEngine(grammar)
        End If
    End Sub

    Private Sub SetSpeechEngine(grammar As Grammar)
        If grammar Is Nothing Then Throw New ArgumentNullException()
        StopSpeechRecognition()
        Dim ri As RecognizerInfo = GetKinectRecognizer(SpeechRecognitionLanguage)
        If Nothing IsNot ri Then
            Me._SpeechEngine = New SpeechRecognitionEngine(ri.Id)

            AddHandler _SpeechEngine.SpeechRecognized, AddressOf SpeechRecognizedHandler
            _SpeechEngine.LoadGrammar(grammar)
            _SpeechEngine.SetInputToAudioStream(_Sensor.AudioSource.Start(), New SpeechAudioFormatInfo(EncodingFormat.Pcm, 16000, 16, 1, 32000, 2, Nothing))
            _SpeechEngine.RecognizeAsync(RecognizeMode.Multiple)
        Else
            Throw New Exception()
        End If
    End Sub

    Public Sub StopSpeechRecognition()
        If Me._SpeechEngine IsNot Nothing Then
            RemoveHandler Me._SpeechEngine.SpeechRecognized, AddressOf SpeechRecognizedHandler
            Me._SpeechEngine.RecognizeAsyncStop()
            Me._SpeechEngine = Nothing
        End If
    End Sub

    Private Shared Function GetKinectRecognizer(language As String) As RecognizerInfo
        For Each recognizer As RecognizerInfo In SpeechRecognitionEngine.InstalledRecognizers()
            Dim value As String = ""
            recognizer.AdditionalInfo.TryGetValue("Kinect", value)
            If String.Compare(value, "true", True) = 0 AndAlso String.Compare(language, recognizer.Culture.Name, True) = 0 Then
                Return recognizer
            End If
        Next recognizer

        Return Nothing
    End Function

    Private Sub SpeechRecognizedHandler(sender As Object, e As SpeechRecognizedEventArgs)
        RaiseEvent SpeechRecognized(Me, e)
    End Sub

#End Region

#Region "Events"
    Public Event GestureRecognized(sender As Object, e As GestureRecognizedEventArgs)
    Public Event SkeletalTracked(sender As Object, e As SkeletalTrackedEventArgs)
    Public Event FrameChanged(sender As Object, e As System.EventArgs)
    Public Event SpeechRecognized As EventHandler(Of SpeechRecognizedEventArgs)
#End Region

#Region "Methods"
    Public Sub Start(mode As KinectManagerMode, features As KinectManagerFeature)
        Me.StartKinectSensor(mode, features)
    End Sub

    Public Sub [Stop]()
        Me.StopKinectSensor()
    End Sub
#End Region

#Region "IDisposable Support"
    Private disposedValue As Boolean ' To detect redundant calls

    ' IDisposable
    Protected Overridable Sub Dispose(disposing As Boolean)
        If Not Me.disposedValue Then
            If disposing Then
                StopSpeechRecognition()
                StopKinectSensor()
                If _Sensor IsNot Nothing Then _Sensor.Dispose()
            End If

            ' TODO: free unmanaged resources (unmanaged objects) and override Finalize() below.
            ' TODO: set large fields to null.
        End If
        Me.disposedValue = True
    End Sub

    ' TODO: override Finalize() only if Dispose(ByVal disposing As Boolean) above has code to free unmanaged resources.
    'Protected Overrides Sub Finalize()
    '    ' Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
    '    Dispose(False)
    '    MyBase.Finalize()
    'End Sub

    ' This code added by Visual Basic to correctly implement the disposable pattern.
    Public Sub Dispose() Implements IDisposable.Dispose
        ' Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub
#End Region

    




End Class
