﻿

Imports System.IO
Imports GadgeteerManager
Imports StateMachineManager.Implementations
Imports KinectManager.Core
Imports Microsoft.Speech.AudioFormat
Imports Microsoft.Kinect
Imports Microsoft.Practices.Unity
Imports System.Threading
Imports Microsoft.Speech.Recognition
Imports Coding4Fun.Kinect.Wpf
Imports System.Reflection
Imports System.Threading.Tasks
Imports System.Configuration

Namespace ViewModels
    Public MustInherit Class KinectViewModelBase
        Inherits ViewModelBase

        Public Sub New(container As IUnityContainer)
            MyBase.New(container)
            If IsInDesignMode Then Return
            TaskCancellationSource = New CancellationTokenSource
            CreateSensor()
            SetSpeechRecognitionEngine()
            SetElevationAngle(15)
        End Sub

        Protected Property CurrentLanguage As String = "it-IT"
        Protected Property Sensor As KinectSensor
        Protected Property SpeechEngine As SpeechRecognitionEngine
        Protected Property KinectMode As KinectMode = KinectMode.Polling
        Protected Property KinectTask As Task
        Protected Property TaskCancellationSource As CancellationTokenSource



        Protected Overrides Sub Dispose(disposing As Boolean)
            MyBase.Dispose(disposing)
            If disposing Then
                If Sensor IsNot Nothing Then
                    If Sensor.IsRunning Then
                        If KinectMode = Core.KinectMode.EventDriven Then
                            RemoveHandler Sensor.AllFramesReady, AddressOf AllFrameReadyHandler
                        Else
                            TaskCancellationSource.Cancel()
                            KinectTask.Wait()
                        End If
                        Sensor.Stop()
                        Sensor.Dispose()
                    End If
                End If
            End If
        End Sub

#Region "Properties"

        Private _CurrentSkeleton As Skeleton
        Public Property CurrentSkeleton As Skeleton
            Get
                Return _CurrentSkeleton
            End Get
            Set(value As Skeleton)
                Me._CurrentSkeleton = value
                NotifyAllPropertyChanged()
            End Set
        End Property

        Public ReadOnly Property IsCurrentSkeletonTracked As Boolean
            Get
                Return CurrentSkeleton IsNot Nothing
            End Get
        End Property

        Private _CurrentColorImage As BitmapSource
        Public Property CurrentColorImage As BitmapSource
            Get
                Return Me._CurrentColorImage
            End Get
            Set(value As BitmapSource)
                Me._CurrentColorImage = value
                OnPropertyChanged("CurrentColorImage")
            End Set
        End Property

        Private _CurrentDepthImage As BitmapSource
        Public Property CurrentDepthImage As BitmapSource
            Get
                Return Me._CurrentDepthImage
            End Get
            Set(value As BitmapSource)
                Me._CurrentDepthImage = value
                OnPropertyChanged("CurrentDepthImage")
            End Set
        End Property

        Private _HeadX As Double
        Public Property HeadX As Double
            Get
                Return _HeadX
            End Get
            Set(value As Double)
                _HeadX = value
                OnPropertyChanged("HeadX")
            End Set
        End Property

        Private _HeadY As Double
        Public Property HeadY As Double
            Get
                Return _HeadY
            End Get
            Set(value As Double)
                _HeadY = value
                OnPropertyChanged("HeadY")
            End Set
        End Property

        Private _HandRightX As Double
        Public Property HandRightX As Double
            Get
                Return _HandRightX
            End Get
            Set(value As Double)
                _HandRightX = value
                OnPropertyChanged("HandRightX")
            End Set
        End Property

        Private _HandRightY As Double
        Public Property HandRightY As Double
            Get
                Return _HandRightY
            End Get
            Set(value As Double)
                _HandRightY = value
                OnPropertyChanged("HandRightY")
            End Set
        End Property

        Private _HandLeftX As Double
        Public Property HandLeftX As Double
            Get
                Return _HandLeftX
            End Get
            Set(value As Double)
                _HandLeftX = value
                OnPropertyChanged("HandLeftX")
            End Set
        End Property

        Private _HandLeftY As Double
        Public Property HandLeftY As Double
            Get
                Return _HandLeftY
            End Get
            Set(value As Double)
                _HandLeftY = value
                OnPropertyChanged("HandLeftY")
            End Set
        End Property

        Public ReadOnly Property ElevationAngle As Integer
            Get
                If Sensor IsNot Nothing AndAlso Sensor.IsRunning Then
                    Return Sensor.ElevationAngle
                Else
                    Return Nothing
                End If
            End Get
        End Property

        Private _CurrentOperation As String
        Public Property CurrentOperation As String
            Get
                Return Me._CurrentOperation
            End Get
            Set(value As String)
                If Me._CurrentOperation <> value Then
                    Me._CurrentOperation = value
                    OnPropertyChanged("CurrentOperation")
                End If
            End Set
        End Property

#End Region

#Region "Kinect"

#Region "Creation Sensor and Speech Engine"
        Protected Sub CreateSensor()
            If KinectSensor.KinectSensors.Any() Then
                Sensor = KinectSensor.KinectSensors.FirstOrDefault(Function(s) s.Status = KinectStatus.Connected)
                If Sensor IsNot Nothing Then
                    Dim parameters = New TransformSmoothParameters() With {.Smoothing = 0.5F,
                                                                           .Correction = 0.5F,
                                                                           .Prediction = 0.5F,
                                                                           .JitterRadius = 0.05F,
                                                                           .MaxDeviationRadius = 0.04F}
                    Sensor.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30)
                    Sensor.DepthStream.Enable(DepthImageFormat.Resolution320x240Fps30)
                    Sensor.SkeletonStream.Enable(parameters)
                    Try
                        Sensor.Start()
                    Catch ex As Exception

                    End Try
                    If KinectMode = Core.KinectMode.EventDriven Then
                        AddHandler Sensor.AllFramesReady, AddressOf AllFrameReadyHandler
                    Else
                        KinectTask = Task.Run(AddressOf KinectTaskHandler, TaskCancellationSource.Token)
                    End If
                End If
            End If
        End Sub

        Protected Sub SetSpeechRecognitionEngine()
            If Sensor IsNot Nothing AndAlso Sensor.IsRunning Then
                Sensor.AudioSource.ManualBeamAngle = 0
                Dim ri As RecognizerInfo = GetKinectRecognizer(CurrentLanguage)
                If Nothing IsNot ri Then
                    Me.SpeechEngine = New SpeechRecognitionEngine(ri.Id)
                    LoadGrammar(CurrentLanguage)
                    AddHandler SpeechEngine.SpeechRecognized, AddressOf SpeechRecognized
                Else
                    Throw New Exception()
                End If
            End If
        End Sub
#End Region

#Region "Gestures"
        Protected MustOverride Sub ManageGestures()
        Protected MustOverride Sub SkeletonNotTracked()
#End Region

#Region "Color, Depth & Skeleton Tracking"
        Protected Skeletons() As Skeleton = New Skeleton(5) {}

        Protected Sub AllFrameReadyHandler(sender As Object, e As AllFramesReadyEventArgs)
            Using depthFrame = Sensor.DepthStream.OpenNextFrame(0),
                    colorFrame = Sensor.ColorStream.OpenNextFrame(0),
                    skeletalFrame = Sensor.SkeletonStream.OpenNextFrame(0)
                ManageFrames(colorFrame, depthFrame, skeletalFrame, False)
            End Using
        End Sub

        Protected Sub KinectTaskHandler()
            While Not TaskCancellationSource.Token.IsCancellationRequested
                Using depthFrame = Sensor.DepthStream.OpenNextFrame(0),
                    colorFrame = Sensor.ColorStream.OpenNextFrame(0),
                    skeletalFrame = Sensor.SkeletonStream.OpenNextFrame(0)

                    ManageFrames(colorFrame, depthFrame, skeletalFrame, True)
                End Using
                Task.Delay(25)
            End While
        End Sub

        Protected Sub ManageFrames(colorFrame As ColorImageFrame, depthFrame As DepthImageFrame, skeletalFrame As SkeletonFrame, useFreeze As Boolean)
            If colorFrame IsNot Nothing Then
                Dim image = colorFrame.ToBitmapSource()
                If useFreeze Then image.Freeze()
                Me.CurrentColorImage = image
            End If

            If depthFrame IsNot Nothing Then
                Dim image = depthFrame.ToBitmapSource()
                If useFreeze Then image.Freeze()
                Me.CurrentDepthImage = image
            End If

            If skeletalFrame IsNot Nothing Then
                skeletalFrame.CopySkeletonDataTo(Skeletons)
                GetSkeletonTracked()
                If CurrentSkeleton IsNot Nothing Then
                    ManageGestures()
                Else
                    SkeletonNotTracked()
                End If
            End If
        End Sub

        Protected MustOverride Sub GetSkeletonTracked()

        Protected Sub SetElevationAngle(angle As Integer)
            If Sensor IsNot Nothing AndAlso Sensor.IsRunning Then
                If angle >= Sensor.MinElevationAngle Then
                    If angle <= Sensor.MaxElevationAngle Then
                        Sensor.ElevationAngle = angle
                    Else
                        Sensor.ElevationAngle = Sensor.MaxElevationAngle
                    End If
                Else
                    Sensor.ElevationAngle = Sensor.MinElevationAngle
                End If
                OnPropertyChanged("ElevationAngle")
            End If
        End Sub
#End Region

#Region "Speech Recognition"
        Protected Sub LoadGrammar(language As String)
            Dim appFileInfo = New FileInfo(Assembly.GetExecutingAssembly().Location)
            Dim grammarFile = Path.Combine(appFileInfo.DirectoryName, String.Format("speechgrammar.{0}.xml", language))

            If File.Exists(grammarFile) Then
                Dim g = New Grammar(grammarFile)
                SpeechEngine.LoadGrammar(g)
                SpeechEngine.SetInputToAudioStream(_Sensor.AudioSource.Start(),
                                                   New SpeechAudioFormatInfo(EncodingFormat.Pcm, 16000,
                                                                             16, 1, 32000, 2, Nothing))
                SpeechEngine.RecognizeAsync(RecognizeMode.Multiple)
            Else
                Throw New Exception()
            End If
        End Sub

        Protected Shared Function GetKinectRecognizer(language As String) As RecognizerInfo
            For Each recognizer As RecognizerInfo In SpeechRecognitionEngine.InstalledRecognizers()
                Dim value As String = ""
                recognizer.AdditionalInfo.TryGetValue("Kinect", value)
                If String.Compare(value, "true", True) = 0 _
                    AndAlso String.Compare(language, recognizer.Culture.Name, True) = 0 Then
                    Return recognizer
                End If
            Next recognizer
            Return Nothing
        End Function

        Protected SpeechRecognitionDictionaty As IDictionary(Of String, Action)

        Protected Sub SpeechRecognized(sender As Object, e As SpeechRecognizedEventArgs)
            If e.Result.Confidence > 0.7 Then
                If SpeechRecognitionDictionaty.ContainsKey(e.Result.Semantics.Value.ToString) Then
                    SpeechRecognitionDictionaty(e.Result.Semantics.Value.ToString).Invoke()
                End If
            End If
        End Sub
#End Region

#End Region



    End Class
End Namespace
