﻿Imports System.Reflection
Imports StateMachineManager.Internals
Imports StateMachineManager.Exceptions
Imports StateMachineManager.Interfaces

Namespace Implementations

    ''' <summary>
    ''' Implementazione della macchina a stati.
    ''' </summary>
    ''' <typeparam name="TEntity"></typeparam>
    ''' <typeparam name="TState"></typeparam>
    ''' <typeparam name="TEvent"></typeparam>
    ''' <remarks></remarks>
    Friend Class StateMachine(Of TEntity, TState, TEvent)
        Implements IInnerStateMachine(Of TEntity, TState, TEvent), IInnerStateMachine, IOuterStateMachine(Of TEntity, TState, TEvent)

        Public Sub New()
            _Transition = New Dictionary(Of TransitionKey(Of TState, TEvent), TransitionData(Of TEntity, TState))()
        End Sub

#Region "IOuterStateMachine"
        ''' <summary>
        ''' Aggiunge una transizione senza funzionalità di controllo alla macchina a stati.
        ''' </summary>
        ''' <param name="startState">Stato iniziale della transizione.</param>
        ''' <param name="event">Evento di transizione.</param>
        ''' <param name="endState">Stato finale della transizione.</param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Function AddTransition(ByVal startState As TState, ByVal [event] As TEvent, ByVal endState As TState) As IOuterStateMachine(Of TEntity, TState, TEvent) Implements IOuterStateMachine(Of TEntity, TState, TEvent).AddTransition
            Return Me.AddTransitionInternal(startState, [event], endState, Nothing, Nothing)
        End Function

        ''' <summary>
        ''' Aggiunge una transizione con funzionalità di controllo alla macchina a stati.
        ''' </summary>
        ''' <param name="startState">Stato iniziale della transizione.</param>
        ''' <param name="event">Evento di transizione.</param>
        ''' <param name="endState">Stato finale della transizione.</param>
        ''' <param name="conditionalLamdba">Lambda expression di controllo. La funzione viene richiamata prima del cambio di stato e se restituisce true, il cambiamento viene eseguito.</param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Function AddTransition(ByVal startState As TState, _
                                      ByVal [event] As TEvent, _
                                      ByVal endState As TState, _
                                      ByVal conditionalLamdba As Func(Of TEntity, Boolean)) As IOuterStateMachine(Of TEntity, TState, TEvent) Implements IOuterStateMachine(Of TEntity, TState, TEvent).AddTransition
            If conditionalLamdba Is Nothing Then Throw New ArgumentNullException("ConditionalLambda non può essere null")
            Return Me.AddTransitionInternal(startState, [event], endState, conditionalLamdba, Nothing)
        End Function

        ''' <summary>
        ''' Aggiunge una transizione con funzionalità di controllo alla macchina a stati.
        ''' </summary>
        ''' <param name="startState">Stato iniziale della transizione.</param>
        ''' <param name="event">Evento di transizione.</param>
        ''' <param name="endState">Stato finale della transizione.</param>
        ''' <param name="executeLamdba">Lambda expression eseguita in caso di cambio stato. La sub viene richiamata al cambio di stato.</param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Function AddTransition(ByVal startState As TState, _
                                      ByVal [event] As TEvent, _
                                      ByVal endState As TState, _
                                      ByVal executeLamdba As Action(Of TEntity)) As IOuterStateMachine(Of TEntity, TState, TEvent) Implements IOuterStateMachine(Of TEntity, TState, TEvent).AddTransition
            If executeLamdba Is Nothing Then Throw New ArgumentNullException("ExecuteLambda non può essere null")
            Return Me.AddTransitionInternal(startState, [event], endState, Nothing, executeLamdba)
        End Function


        ''' <summary>
        ''' Aggiunge una transizione con funzionalità di controllo alla macchina a stati.
        ''' </summary>
        ''' <param name="startState">Stato iniziale della transizione.</param>
        ''' <param name="event">Evento di transizione.</param>
        ''' <param name="endState">Stato finale della transizione.</param>
        ''' <param name="conditionalLamdba">Lambda expression di controllo. La funzione viene richiamata prima del cambio di stato e se restituisce true, il cambiamento viene eseguito.</param>
        ''' <param name="executeLamdba">Lambda expression eseguita in caso di cambio stato. La sub viene richiamata al cambio di stato.</param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Function AddTransition(ByVal startState As TState, _
                                      ByVal [event] As TEvent, _
                                      ByVal endState As TState, _
                                      ByVal conditionalLamdba As Func(Of TEntity, Boolean),
                                      ByVal executeLamdba As Action(Of TEntity)) As IOuterStateMachine(Of TEntity, TState, TEvent) Implements IOuterStateMachine(Of TEntity, TState, TEvent).AddTransition
            If conditionalLamdba Is Nothing Then Throw New ArgumentNullException("ConditionalLambda non può essere null")
            If executeLamdba Is Nothing Then Throw New ArgumentNullException("ExecuteLambda non può essere null")
            Return Me.AddTransitionInternal(startState, [event], endState, conditionalLamdba, executeLamdba)
        End Function

        Private Function AddTransitionInternal(ByVal startState As TState, _
                                               ByVal [event] As TEvent, _
                                               ByVal endState As TState, _
                                               ByVal conditionalLamdba As Func(Of TEntity, Boolean),
                                               ByVal executeLamdba As Action(Of TEntity)) As IOuterStateMachine(Of TEntity, TState, TEvent)

            Dim transitionKey = New TransitionKey(Of TState, TEvent)() With {.State = startState, .Event = [event]}
            If Not _Transition.ContainsKey(transitionKey) Then
                Dim transitionData = New TransitionData(Of TEntity, TState)() With {.State = endState,
                                                                                    .ConditionLambda = conditionalLamdba,
                                                                                    .ExecuteLambda = executeLamdba}
                Me._Transition.Add(transitionKey, transitionData)
            End If
            Return Me
        End Function
#End Region

#Region "IInnerStateMachine"
        ''' <summary>
        ''' Changes the state.
        ''' </summary>
        ''' <param name="entity">The entity.</param>
        ''' <param name="event">The event.</param>
        ''' <param name="oldState">The old state.</param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Function ChangeState(ByVal entity As TEntity, _
                                    ByVal [event] As TEvent, _
                                    ByRef oldState As TState) As Boolean Implements IInnerStateMachine(Of TEntity, TState, TEvent).ChangeState
            If entity Is Nothing Then Throw New ArgumentNullException("Entity non può essere null")
            Dim retval = False
            If _PropertyStateInfo IsNot Nothing Then
                Dim tmpOldState = CType(_PropertyStateInfo.GetValue(entity, Nothing), TState)
                Dim transitionKey = New TransitionKey(Of TState, TEvent)() With {.State = tmpOldState, .Event = [event]}
                If _Transition.ContainsKey(transitionKey) Then
                    Dim transitionData = Me._Transition.Item(transitionKey)
                    If transitionData.ConditionLambda Is Nothing OrElse transitionData.ConditionLambda(entity) Then
                        _PropertyStateInfo.SetValue(entity, transitionData.State, Nothing)
                        oldState = tmpOldState
                        If transitionData.ExecuteLambda IsNot Nothing Then
                            transitionData.ExecuteLambda.Invoke(entity)
                        End If
                        retval = True
                    End If
                End If
            Else
                Throw New StateMachineException("Proprietà di stato non definita")
            End If
            Return retval
        End Function

        ''' <summary>
        ''' Imposta la proprietà di stato di una macchina a stati.
        ''' </summary>
        ''' <param name="propertyName">Nome della proprietà.</param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Function SetStateProperty(ByVal propertyName As String) As Boolean Implements IInnerStateMachine(Of TEntity, TState, TEvent).SetStateProperty
            If String.IsNullOrEmpty(propertyName) Then Throw New ArgumentNullException("propertyName non può essere null")
            Dim retVal = True
            Dim propertyInfo = GetType(TEntity).GetProperty(propertyName)
            If propertyInfo IsNot Nothing Then
                If propertyInfo.PropertyType Is GetType(TState) Then
                    _PropertyStateInfo = propertyInfo
                Else
                    retVal = False
                    Throw New StateMachineException("Proprietà di stato con tipo difforme dal tipo dello stato")
                End If
            Else
                retVal = False
                Throw New StateMachineException("Proprietà inesistente")
            End If
            Return retVal
        End Function

        ''' <summary>
        ''' Ripulisce la macchina a stati.
        ''' </summary>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Function Clear() As Boolean Implements IInnerStateMachine(Of TEntity, TState, TEvent).Clear
            Me._Transition.Clear()
            Return True
        End Function
#End Region

        ''' <summary>
        ''' Dictionary che contiene le coppie (stato,Evento) e le relative coppie (stato, lambda) che dfiniscono una transizione.
        ''' </summary>
        ''' <remarks></remarks>
        Private _Transition As Dictionary(Of TransitionKey(Of TState, TEvent), TransitionData(Of TEntity, TState))

        ''' <summary>
        ''' Property info della proprietà contenente lo stato.
        ''' </summary>
        ''' <remarks></remarks>
        Private _PropertyStateInfo As PropertyInfo

    End Class
End Namespace