﻿
Imports System.IO
Imports StateMachineManager.EventArgs
Imports StateMachineManager.Exceptions
Imports StateMachineManager.Interfaces

Namespace Implementations
    ''' <summary>
    ''' Classe per la gestione delle macchine a stati.
    ''' </summary>
    ''' <remarks></remarks>
    Public NotInheritable Class StateMachineController
        Private Sub New()

        End Sub

        Shared Sub New()
            _StateMachineCollection = New List(Of Object)
            _Observers = New List(Of IStateMachineobserver)
        End Sub

        ''' <summary>
        ''' Struttura che memorizza l'insieme delle macchine a stati definite.
        ''' </summary>
        ''' <remarks></remarks>
        Private Shared _StateMachineCollection As ICollection(Of Object)

        ''' <summary>
        ''' Recupera un istanza di IInnerStateMachine.
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Shared Function GetIStateMachine(Of TEntity, TState, TEvent)() As IInnerStateMachine(Of TEntity, TState, TEvent)
            Dim retval As IInnerStateMachine(Of TEntity, TState, TEvent) = Nothing
            If _StateMachineCollection IsNot Nothing Then
                retval = (From i In _StateMachineCollection _
                         Where TypeOf i Is IInnerStateMachine(Of TEntity, TState, TEvent) _
                         Select CType(i, IInnerStateMachine(Of TEntity, TState, TEvent))).FirstOrDefault()
            End If
            Return retval
        End Function

        ''' <summary>
        ''' Crea una macchina a stati (istanza di tipo StateMachine).
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Shared Function CreateStateMachine(Of TEntity, TState, TEvent)() As StateMachine(Of TEntity, TState, TEvent)
            Dim stateMachine = New StateMachine(Of TEntity, TState, TEvent)()
            _StateMachineCollection.Add(stateMachine)
            Return stateMachine
        End Function

        ''' <summary>
        ''' Crea una istanza di IInnerStateMachine.
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Shared Function CreateInnerStateMachine(Of TEntity, TState, TEvent)() As IInnerStateMachine(Of TEntity, TState, TEvent)
            Return CType(CreateStateMachine(Of TEntity, TState, TEvent)(), IInnerStateMachine(Of TEntity, TState, TEvent))
        End Function

        ''' <summary>
        ''' Crea una istanza di IOuterStateMachine.
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Shared Function CreateOuterStateMachine(Of TEntity, TState, TEvent)() As IOuterStateMachine(Of TEntity, TState, TEvent)
            Return CType(CreateStateMachine(Of TEntity, TState, TEvent)(), IOuterStateMachine(Of TEntity, TState, TEvent))
        End Function

        ''' <summary>
        ''' Crea una macchina a stati.
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <param name="stateProperty">Nome della proprietà di stato.</param>
        ''' <exception cref="Exceptions.StateMachineException">Se la macchina a stati non uò essere creata.</exception>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Shared Function CreateMachine(Of TEntity, TState, TEvent)(ByVal stateProperty As String) As IOuterStateMachine(Of TEntity, TState, TEvent)
            WriteToLog("Inizio creazione macchina a stati TEntity={0} TState={1} TEvent={2} StateProperty={3}", GetType(TEntity), GetType(TState), GetType(TEvent), stateProperty)
            Dim stateMachine = CreateStateMachine(Of TEntity, TState, TEvent)()
            If stateMachine IsNot Nothing Then
                stateMachine.SetStateProperty(stateProperty)
            Else
                Throw New StateMachineException()
            End If
            WriteToLog("Fine creazione macchina a stati TEntity={0} TState={1} TEvent={2} StateProperty={3}", GetType(TEntity), GetType(TState), GetType(TEvent), stateProperty)
            Return stateMachine
        End Function

        ''' <summary>
        ''' Esegue un cambio di stato per una macchina a stati.
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <param name="entity">Entità su cui eseguire il cambio di stato.</param>
        ''' <param name="event">Evento che provoca il cambio di stato.</param>
        ''' <exception cref="Exceptions.StateMachineException">Se la macchina a stati non esiste.</exception>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Shared Function ChangeState(Of TEntity, TState, TEvent)(ByVal entity As TEntity, ByVal [event] As TEvent) As Boolean
            WriteToLog("Inizio cambio stato TEntity={0} TState={1} TEvent={2} Entity={3} Event={4}", GetType(TEntity), GetType(TState), GetType(TEvent), entity, [event])
            Dim retval As Boolean = False
            Dim stateMachine = GetIStateMachine(Of TEntity, TState, TEvent)()
            If stateMachine IsNot Nothing Then
                If OnStateChanging(Of TEntity, TState, TEvent)(entity, [event]) Then
                    Dim oldState As TState
                    retval = stateMachine.ChangeState(entity, [event], oldState)
                    If retval Then
                        OnStateChanged(Of TEntity, TState, TEvent)(entity, [event], oldState)
                    Else
                        OnStateChangeError(Of TEntity, TState, TEvent)(entity, [event])
                    End If
                End If
            Else
                Throw New StateMachineException()
            End If
            WriteToLog("Fine cambio stato TEntity={0} TState={1} TEvent={2} Entity={3} Event={4} Risultato={5}", GetType(TEntity), GetType(TState), GetType(TEvent), entity, [event], retval)
            Return retval
        End Function

        ''' <summary>
        ''' Rimuove una macchina a stati.
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <exception cref="Exceptions.StateMachineException">Se la macchina a stati non esiste.</exception>
        ''' <remarks></remarks>
        Public Shared Sub RemoveStateMachine(Of TEntity, TState, TEvent)()
            Dim stateMachine = GetIStateMachine(Of TEntity, TState, TEvent)()
            If stateMachine IsNot Nothing Then
                stateMachine.Clear()
                _StateMachineCollection.Remove(stateMachine)
            Else
                Throw New StateMachineException()
            End If
        End Sub

        ''' <summary>
        ''' Elimina le transizioni di una macchina a stati.
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <exception cref="Exceptions.StateMachineException">Se la macchina a stati non esiste.</exception>
        ''' <remarks></remarks>
        Public Shared Sub ClearStateMachine(Of TEntity, TState, TEvent)()
            Dim stateMachine = GetIStateMachine(Of TEntity, TState, TEvent)()
            If stateMachine IsNot Nothing Then
                stateMachine.Clear()
            Else
                Throw New StateMachineException()
            End If
        End Sub

        ''' <summary>
        ''' Rimuove tutte le macchine a stati.
        ''' </summary>
        ''' <remarks></remarks>
        Public Shared Sub RemoveAllStateMachines()
            For Each sm In _StateMachineCollection.OfType(Of IInnerStateMachine)()
                sm.Clear()
            Next
            _StateMachineCollection.Clear()
        End Sub

        ''' <summary>
        ''' Gestione dell'evento StateChanging.
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <param name="entity"></param>
        ''' <param name="event"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Shared Function OnStateChanging(Of TEntity, TState, TEvent)(ByVal entity As TEntity, ByVal [event] As TEvent) As Boolean
            WriteToLog("Inizio OnStateChanging TEntity={0} TState={1} TEvent={2} Entity={3} Event={4}", GetType(TEntity), GetType(TState), GetType(TEvent), entity, [event])
            Dim changingArg = CreateStateChangingEventArgs(Of TEntity, TState, TEvent)(entity, [event])
            RaiseEvent StateChanging(changingArg)
            Dim cancel = False
            For i = 0 To _Observers.Count - 1
                _Observers(i).StateChanging(entity, _
                                            GetType(TEntity), _
                                            [event], _
                                            GetType(TEvent), _
                                            GetType(TState), _
                                            cancel)
                If cancel Then Exit For
            Next
            WriteToLog("Fine OnStateChanging TEntity={0} TState={1} TEvent={2} Entity={3} Event={4}", GetType(TEntity), GetType(TState), GetType(TEvent), entity, [event])
            Return Not changingArg.Cancel
        End Function

        ''' <summary>
        ''' Gestione dell'evento StateChanged.
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <param name="entity"></param>
        ''' <param name="event"></param>
        ''' <param name="oldState"></param>
        ''' <remarks></remarks>
        Private Shared Sub OnStateChanged(Of TEntity, TState, TEvent)(ByVal entity As TEntity, ByVal [event] As TEvent, ByVal oldState As TState)
            WriteToLog("Inizio OnStateChanged TEntity={0} TState={1} TEvent={2} Entity={3} Event={4}", GetType(TEntity), GetType(TState), GetType(TEvent), entity, [event])
            Dim changedArg = CreateStateChangedEventArgs(Of TEntity, TState, TEvent)(entity, [event])
            changedArg.OldState = oldState
            RaiseEvent StateChanged(changedArg)
            Dim cancel = False
            For i = 0 To _Observers.Count - 1
                _Observers(i).StateChanged(entity, GetType(TEntity), [event], GetType(TEvent), oldState, GetType(TState))
                If cancel Then Exit For
            Next
            WriteToLog("Fine OnStateChanged TEntity={0} TState={1} TEvent={2} Entity={3} Event={4}", GetType(TEntity), GetType(TState), GetType(TEvent), entity, [event])
        End Sub

        ''' <summary>
        ''' Gestione dell'evento StateChangeError.
        ''' </summary>
        ''' <typeparam name="TEntity">Tipo dell'entità gestita dalla macchina a stati.</typeparam>
        ''' <typeparam name="TState">Tipo degli stati gestiti dalla macchina a stati.</typeparam>
        ''' <typeparam name="TEvent">Tipo degli eventi gestiti dalla macchina a stati.</typeparam>
        ''' <param name="entity"></param>
        ''' <param name="event"></param>
        ''' <remarks></remarks>
        Private Shared Sub OnStateChangeError(Of TEntity, TState, TEvent)(ByVal entity As TEntity, ByVal [event] As TEvent)
            WriteToLog("Inizio OnStateChangeError TEntity={0} TState={1} TEvent={2} Entity={3} Event={4}", GetType(TEntity), GetType(TState), GetType(TEvent), entity, [event])
            Dim changeErrorArg = CreateStateChangeErrorEventArgs(Of TEntity, TState, TEvent)(entity, [event])
            RaiseEvent StateChangeError(changeErrorArg)
            Dim cancel = False
            For i = 0 To _Observers.Count - 1
                _Observers(i).StateChangeError(entity, GetType(TEntity), [event], GetType(TEvent), GetType(TState))
                If cancel Then Exit For
            Next
            WriteToLog("Fine OnStateChangeError TEntity={0} TState={1} TEvent={2} Entity={3} Event={4}", GetType(TEntity), GetType(TState), GetType(TEvent), entity, [event])
        End Sub

#Region "Events"
        ''' <summary>
        ''' 
        ''' </summary>
        ''' <param name="e"></param>
        ''' <remarks></remarks>
        Public Shared Event StateChanging(ByVal e As StateChangingEventArgs)

        ''' <summary>
        ''' 
        ''' </summary>
        ''' <param name="e"></param>
        ''' <remarks></remarks>
        Public Shared Event StateChanged(ByVal e As StateChangedEventArgs)

        ''' <summary>
        ''' 
        ''' </summary>
        ''' <param name="e"></param>
        ''' <remarks></remarks>
        Public Shared Event StateChangeError(ByVal e As StateChangeErrorEventArgs)
#End Region

#Region "Event Manage Methods"
        ''' <summary>
        ''' 
        ''' </summary>
        ''' <typeparam name="TEntity"></typeparam>
        ''' <typeparam name="TState"></typeparam>
        ''' <typeparam name="TEvent"></typeparam>
        ''' <param name="entity"></param>
        ''' <param name="event"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Shared Function CreateStateChangingEventArgs(Of TEntity, TState, TEvent)(ByVal entity As TEntity, ByVal [event] As TEvent) As StateChangingEventArgs
            Dim arg = New StateChangingEventArgs(CreateStateChangeEventArgsBase(Of TEntity, TState, TEvent)(entity, [event]))
            Return arg
        End Function

        ''' <summary>
        ''' 
        ''' </summary>
        ''' <typeparam name="TEntity"></typeparam>
        ''' <typeparam name="TState"></typeparam>
        ''' <typeparam name="TEvent"></typeparam>
        ''' <param name="entity"></param>
        ''' <param name="event"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Shared Function CreateStateChangedEventArgs(Of TEntity, TState, TEvent)(ByVal entity As TEntity, ByVal [event] As TEvent) As StateChangedEventArgs
            Dim arg = New StateChangedEventArgs(CreateStateChangeEventArgsBase(Of TEntity, TState, TEvent)(entity, [event]))
            Return arg
        End Function

        ''' <summary>
        ''' 
        ''' </summary>
        ''' <typeparam name="TEntity"></typeparam>
        ''' <typeparam name="TState"></typeparam>
        ''' <typeparam name="TEvent"></typeparam>
        ''' <param name="entity"></param>
        ''' <param name="event"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Shared Function CreateStateChangeErrorEventArgs(Of TEntity, TState, TEvent)(ByVal entity As TEntity, ByVal [event] As TEvent) As StateChangeErrorEventArgs
            Dim arg = New StateChangeErrorEventArgs(CreateStateChangeEventArgsBase(Of TEntity, TState, TEvent)(entity, [event]))
            Return arg
        End Function

        ''' <summary>
        ''' 
        ''' </summary>
        ''' <typeparam name="TEntity"></typeparam>
        ''' <typeparam name="TState"></typeparam>
        ''' <typeparam name="TEvent"></typeparam>
        ''' <param name="entity"></param>
        ''' <param name="event"></param>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Private Shared Function CreateStateChangeEventArgsBase(Of TEntity, TState, TEvent)(ByVal entity As TEntity, ByVal [event] As TEvent) As StateChangeEventArgsBase
            Dim arg = New StateChangeEventArgsBase()
            With arg
                .Entity = entity
                .EntityType = GetType(TEntity)
                .Event = [event]
                .EventType = GetType(TEvent)
                .StateType = GetType(TState)
            End With
            Return arg
        End Function
#End Region

#Region "Gestione Text Writer"

        Private Shared Sub WriteToLog(ByVal message As String)
            If message Is Nothing Then Throw New ArgumentNullException("Message deve essere not null")
            If _Log IsNot Nothing Then
                _Log.WriteLine("{0:yyyy-MM-dd HH:mm:ss} - {1}", DateTime.Now, message)
            End If
        End Sub

        Private Shared Sub WriteToLog(ByVal formatString As String, ByVal ParamArray args() As Object)
            WriteToLog(String.Format(formatString, args))
        End Sub

        Private Shared _Log As TextWriter

        ''' <summary>
        ''' 
        ''' </summary>
        ''' <value></value>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Shared Property Log() As TextWriter
            Get
                Return _Log
            End Get
            Set(ByVal value As TextWriter)
                If value Is Nothing Then Throw New ArgumentNullException("Il text writer non può essere null")
                _Log = value
            End Set
        End Property
#End Region

#Region "Gestione IStateMachineObserver"
        Private Shared _Observers As List(Of IStateMachineobserver)

        ''' <summary>
        ''' 
        ''' </summary>
        ''' <param name="observer"></param>
        ''' <remarks></remarks>
        Public Shared Sub AddObserver(ByVal observer As IStateMachineObserver)
            If observer Is Nothing Then Throw New ArgumentNullException("Observer non può essere null")
            _Observers.Add(observer)
        End Sub

        ''' <summary>
        ''' 
        ''' </summary>
        ''' <param name="observer"></param>
        ''' <remarks></remarks>
        Public Shared Sub RemoveObserver(ByVal observer As IStateMachineObserver)
            If observer Is Nothing Then Throw New ArgumentNullException("Observer non può essere null")
            _Observers.Remove(observer)
        End Sub

        ''' <summary>
        ''' 
        ''' </summary>
        ''' <remarks></remarks>
        Public Shared Sub RemoveAllObserver()
            _Observers.Clear()
        End Sub

        ''' <summary>
        ''' 
        ''' </summary>
        ''' <value></value>
        ''' <returns></returns>
        ''' <remarks></remarks>
        Public Shared ReadOnly Property Observers() As IEnumerable(Of IStateMachineObserver)
            Get
                Return _Observers
            End Get
        End Property
#End Region

    End Class

End Namespace